////////////////////////////////////////////////////
// CLuaEngineManager.cpp
//
// Purpose: Manages the Lua engine
//
// File History:
//	- 8/13/07 : File created - Kevin Kirst
////////////////////////////////////////////////////

#include "CLuaEngineManager.h"
#include "scripts.h"
#include "engine.h"
#include "plugin.h"

// Script objects
#include "CScriptObjectCommands.h"
#include "CScriptObjectLevel.h"
#include "CScriptObjectSystem.h"
#include "CScriptObjectPlayer.h"

// Set up exposure
IScriptSystem *Lua = NULL;
CLuaEngineManager CLuaEngineManager::m_Instance;
CLuaEngineManager *LuaManager = &CLuaEngineManager::GetInstance();

////////////////////////////////////////////////////
CLuaEngineManager::CLuaEngineManager(void)
{
	m_pUpdateObject = NULL;
	m_pSystemScriptObject = NULL;
	m_pCommandsScriptObject = NULL;
	m_pLevelScriptObject = NULL;

	// Get instance
	m_pSS = GetScriptSystem();
	Lua = m_pSS;

	// Set error callback
	SetErrorCallback(CLuaEngineManager::OnLuaError);
}

////////////////////////////////////////////////////
CLuaEngineManager::~CLuaEngineManager(void)
{
	// Release templates
	CScriptObjectCommands::ReleaseTemplate();
	CScriptObjectLevel::ReleaseTemplate();
	CScriptObjectSystem::ReleaseTemplate();
	CScriptObjectPlayer::ReleaseTemplate();

	// Delete script objects
	SAFE_DELETE(m_pCommandsScriptObject);
	SAFE_DELETE(m_pLevelScriptObject);
	SAFE_DELETE(m_pCommandsScriptObject);

	// Release instance
	SAFE_RELEASE(m_pSS);
}

////////////////////////////////////////////////////
void CLuaEngineManager::Initialize(void)
{
	// Initialize script templates
	CScriptObjectCommands::InitializeTemplate(m_pSS);
	CScriptObjectLevel::InitializeTemplate(m_pSS);
	CScriptObjectSystem::InitializeTemplate(m_pSS);
	CScriptObjectPlayer::InitializeTemplate(m_pSS);

	// Create script objects
	m_pCommandsScriptObject = new CScriptObjectCommands(m_pSS);
	m_pLevelScriptObject = new CScriptObjectLevel;
	m_pSystemScriptObject = new CScriptObjectSystem(m_pSS);
}

////////////////////////////////////////////////////
void CLuaEngineManager::GetScriptPath(char const* szScript, char *szPath)
{
	strcpy(szPath, SCRIPT_PATH_BASE);
	strcat(szPath, szScript);
}

////////////////////////////////////////////////////
std::string CLuaEngineManager::GetScriptPath(char const* szScript)
{
	std::string szTemp = SCRIPT_PATH_BASE;
	szTemp += szScript;
	return szTemp;
}

////////////////////////////////////////////////////
void CLuaEngineManager::OnLuaError(char const* szError)
{
	// Shove it out to the console
	Console_Output("%s\n",szError);
}

////////////////////////////////////////////////////
CLuaEngineManager& CLuaEngineManager::GetInstance(void)
{
	return m_Instance;
}

////////////////////////////////////////////////////
CLuaEngineManager::operator IScriptSystem*(void)
{
	return m_pSS;
}

////////////////////////////////////////////////////
void CLuaEngineManager::OnLevelLoaded(char const* szLevel)
{
	// Call old script's unloaded
	if (true == m_pSS->BeginCall(m_szCurrLevel.c_str(), "OnUnloaded"))
	{
		m_pSS->PushArgument(m_pLevelScriptObject->GetScriptObject());
		m_pSS->EndCall();
	}

	// Create new update object
	Vector3 pos; memset(&pos, 0, sizeof(Vector3));
	m_pUpdateObject = Commands->Create_Object("Invisible_Object", pos);
	Attach_Script_Once(m_pUpdateObject, "LUA_INTERNAL_UPDATE", "");
	m_dwLastTick = GetTickCount();

	// Strip extension
	m_szCurrLevel = szLevel;
	m_szCurrLevel = m_szCurrLevel.substr(0, m_szCurrLevel.find("."));
	if (m_szCurrLevel.at(1) == '&')
		m_szCurrLevel[1] = 'N';

	// Load the map script
	std::string szLevelFullPath = SCRIPT_PATH_MAPS + m_szCurrLevel + ".lua";
	if (false == m_pSS->ReloadScript(szLevelFullPath.c_str()))
	{
		// Try common map
		if (false == m_pSS->ReloadScript(GetScriptPath("Maps\\DefaultMap.lua").c_str()))
		{
			m_pSS->RaiseError("Could not attach a level script to loaded map \'%s\'", szLevel);
			m_szCurrLevel.clear();
		}
		else
			m_szCurrLevel = "DefaultMap";
	}
	if (true == m_szCurrLevel.empty()) return;

	// Reattach the level script object
	m_pLevelScriptObject->SetLevel(m_pSS, m_szCurrLevel.c_str());

	// Call its loaded
	if (true == m_pSS->BeginCall(m_szCurrLevel.c_str(), "OnLoaded"))
	{
		m_pSS->PushArgument(m_pLevelScriptObject->GetScriptObject());
		m_pSS->EndCall();
	}
}

////////////////////////////////////////////////////
void CLuaEngineManager::OnObjectCreated(GameObject *pObj)
{
	if (NULL == m_pLevelScriptObject) return;

	// Get the map's preset registry
	IScriptObject *pMapObj = m_pLevelScriptObject->GetScriptObject();
	if (NULL == pMapObj) return;
	IScriptObject *pPresetList = m_pSS->CreateEmptyObject();
	if (true == pMapObj->GetValue("PresetRegistry", pPresetList))
	{
		// Search for entry in preset registry
		IScriptObject *pPresetEntry = m_pSS->CreateEmptyObject();
		if (true == pPresetList->GetValue(Commands->Get_Preset_Name(pObj), pPresetEntry))
		{
			// Attach a Lua Preset script to it and load the script
			Console_Output("Preset entry found for prest %s\n", Commands->Get_Preset_Name(pObj));
		}
		pPresetEntry->Release();
	}
	pPresetList->Release();
}

////////////////////////////////////////////////////
void CLuaEngineManager::OnPurchase(GameObject *pPurchaser, unsigned int nCost,
								   unsigned int nPresetID)
{
	// Get player ID
	if (NULL == pPurchaser) return;
	int nPlayerID = Get_Player_ID(pPurchaser);

	// Get player entry
	IScriptObject *pPlayerSO = NULL;
	PlayerMap::iterator itPlayer = m_PlayerSOMap.find(nPlayerID);
	if (itPlayer != m_PlayerSOMap.end())
		pPlayerSO = itPlayer->second->GetScriptObject();
	if (NULL == pPlayerSO) return;

	// Inform map
	if (NULL != m_pLevelScriptObject && false == m_szCurrLevel.empty())
	{
		if (true == m_pSS->BeginCall(m_szCurrLevel.c_str(), "OnPurchase"))
		{
			m_pSS->PushArgument(m_pLevelScriptObject->GetScriptObject());
			m_pSS->PushArgument(pPlayerSO);
			m_pSS->PushArgument(nCost);
			m_pSS->PushArgument(Get_Definition_Name(nPresetID));
			m_pSS->EndCall();
		}
	}

	// Inform player script
	if (true == m_pSS->BeginCall("Player", "OnPurchase"))
	{
		m_pSS->PushArgument(pPlayerSO);
		m_pSS->PushArgument(nCost);
		m_pSS->PushArgument(Get_Definition_Name(nPresetID));
		m_pSS->EndCall();
	}
}

////////////////////////////////////////////////////
void CLuaEngineManager::OnTick(void)
{
	// Calculate change in time
	DWORD dwCurrTick = GetTickCount();
	float fDT = float(dwCurrTick - m_dwLastTick)/1000.0f;
	m_dwLastTick = dwCurrTick;

	// Update level
	if (NULL != m_pLevelScriptObject)
		m_pLevelScriptObject->OnTick(fDT);

	// Update all players
	for (PlayerMap::iterator itPlayer = m_PlayerSOMap.begin(); itPlayer != m_PlayerSOMap.end();
		itPlayer++)
	{
		itPlayer->second->OnTick(fDT);
	}
}

////////////////////////////////////////////////////
void CLuaEngineManager::OnChatMessage(int nID, int nType, char const* szMsg)
{
	// Check type
	if (nType != CHAT_PUBLIC && nType != CHAT_PRIVATE && nType  != CHAT_TEAM) return;

	// Get speaker
	IScriptObject *pPlayerSO = NULL;
	PlayerMap::iterator itPlayer = m_PlayerSOMap.find(nID);
	if (itPlayer != m_PlayerSOMap.end())
		pPlayerSO = itPlayer->second->GetScriptObject();

	// Get team (if team messgae)
	int nTeam = -1;
	if (nType == CHAT_TEAM)
	{
		nTeam = Get_Team(nID);
	}

	if (szMsg[0] != '!')
	{
		// Go through OnChat
		if (false == m_szCurrLevel.empty() && true == m_pSS->BeginCall(m_szCurrLevel.c_str(), "OnChat"))
		{
			m_pSS->PushArgument(m_pLevelScriptObject->GetScriptObject());

			// channel
			m_pSS->PushArgument(nType);
			// team
			if (-1 == nTeam) m_pSS->PushNullArgument();
			else m_pSS->PushArgument(nTeam);
			// speaker
			if (NULL != pPlayerSO) m_pSS->PushArgument(pPlayerSO);
			else m_pSS->PushNullArgument();
			// message
			m_pSS->PushArgument(szMsg);

			m_pSS->EndCall();
		}
	}
	else
	{
		// Separate into command and args part
		IScriptObject *pArgTable = m_pSS->CreateObject();
		std::string szCompleteMsg = szMsg, szCommand;
		std::string::size_type nNextSpace = szCompleteMsg.find(' ', 1), nLastSpace;
		if (std::string::npos == nNextSpace)
		{
			// Use all of it except first letter
			szCommand = szCompleteMsg.substr(1, szCompleteMsg.size()-1);
		}
		else
		{
			// Use the part up to the first space
			szCommand = szCompleteMsg.substr(1, nNextSpace-1);

			// Fill arguments
			std::string szArg;
			do
			{
				szArg.clear();
				nLastSpace = nNextSpace;
				nNextSpace = szCompleteMsg.find(' ', nNextSpace+1);
				if (std::string::npos == nNextSpace)
				{
					// Get rest and use it
					szArg = szCompleteMsg.substr(nLastSpace+1, szCompleteMsg.size()-nLastSpace-1);
				}
				else
				{
					// Use part in-between spaces
					szArg = szCompleteMsg.substr(nLastSpace+1, nNextSpace-nLastSpace-1);
				}
				pArgTable->PushValue(szArg.c_str());
			} while (std::string::npos != nNextSpace);
		}

		// Lower case command
		std::string::size_type nSize = szCommand.size();
		for (std::string::size_type i = 0; i < nSize; i++)
		{
			szCommand[i] = (std::string::value_type)tolower((int)szCommand[i]);
		}

		// Go through OnCommand
		if (false == m_szCurrLevel.empty() && true == m_pSS->BeginCall(m_szCurrLevel.c_str(), "OnCommand"))
		{
			m_pSS->PushArgument(m_pLevelScriptObject->GetScriptObject());

			// player
			if (NULL != pPlayerSO) m_pSS->PushArgument(pPlayerSO);
			else m_pSS->PushNullArgument();
			// command
			m_pSS->PushArgument(szCommand);
			// args
			if (pArgTable->Size() == 0) m_pSS->PushNullArgument();
			else m_pSS->PushArgument(pArgTable);
			// channel
			m_pSS->PushArgument(nType);
			// team
			if (-1 == nTeam) m_pSS->PushNullArgument();
			else m_pSS->PushArgument(nTeam);

			m_pSS->EndCall();
		}

		// Clean up
		pArgTable->Release();
	}
}

////////////////////////////////////////////////////
void CLuaEngineManager::OnHostMessage(int nType, char const* szMsg)
{
	// Check type
	if (nType != CHAT_PUBLIC && nType != CHAT_PRIVATE && nType  != CHAT_TEAM) return;

	// Notify
	if (false == m_szCurrLevel.empty() && true == m_pSS->BeginCall(m_szCurrLevel.c_str(), "OnHostMessage"))
	{
		m_pSS->PushArgument(m_pLevelScriptObject->GetScriptObject());
		m_pSS->PushArgument(nType);
		m_pSS->PushArgument(szMsg);
		m_pSS->EndCall();
	}
}

////////////////////////////////////////////////////
void CLuaEngineManager::OnPlayerJoined(int nPlayerID)
{
	// If player already here, force them out
	PlayerMap::iterator itPlayer = m_PlayerSOMap.find(nPlayerID);
	if (itPlayer != m_PlayerSOMap.end())
	{
		// Call OnLeave
		if (true == m_pSS->BeginCall("Player", "OnLeave"))
		{
			m_pSS->PushArgument(itPlayer->second->GetScriptObject());
			m_pSS->EndCall();
		}

		// Clear it
		SAFE_DELETE(itPlayer->second);
		m_PlayerSOMap.erase(itPlayer);
	}

	// Make sure player script is loaded
	std::string szPlayerScript = std::string(SCRIPT_PATH_ENTITIES) + "Players\\Player.lua";
	m_pSS->LoadScript(szPlayerScript.c_str());

	// Create entry
	CScriptObjectPlayer *pSOPlayer = new CScriptObjectPlayer;
	pSOPlayer->SetPlayer(m_pSS, nPlayerID);

	// Call OnJoined
	if (true == m_pSS->BeginCall("Player", "OnJoined"))
	{
		m_pSS->PushArgument(pSOPlayer->GetScriptObject());
		int nRet = 0;
		m_pSS->EndCall(nRet);

		// Kick player out if nRet is 1
		if (1 == nRet)
		{
			char szKickCommand[10] = {0,};
			sprintf(szKickCommand, "kick %d", nPlayerID);
			Console_Input(szKickCommand);
			SAFE_DELETE(pSOPlayer);
			return;
		}
	}
	
	// Add entry to map
	m_PlayerSOMap[nPlayerID] = pSOPlayer;
}

////////////////////////////////////////////////////
void CLuaEngineManager::OnPlayerLeft(int nPlayerID)
{
	// Inform script and remove their entry
	PlayerMap::iterator itPlayer = m_PlayerSOMap.find(nPlayerID);
	if (itPlayer != m_PlayerSOMap.end())
	{
		// Call OnLeave
		if (true == m_pSS->BeginCall("Player", "OnLeave"))
		{
			m_pSS->PushArgument(itPlayer->second->GetScriptObject());
			m_pSS->EndCall();
		}

		// Clear it
		SAFE_DELETE(itPlayer->second);
		m_PlayerSOMap.erase(itPlayer);
	}
}

////////////////////////////////////////////////////
////////////////////////////////////////////////////

////////////////////////////////////////////////////
void LUA_INTERNAL_UPDATE::Created(GameObject *pObj)
{
	// Start update tick
	Commands->Start_Timer(pObj, this, LUAM_UPDATE_TICK_RATE, 1);
}

////////////////////////////////////////////////////
void LUA_INTERNAL_UPDATE::Timer_Expired(GameObject *pObj, int nTimerID)
{
	if (1 == nTimerID)
	{
		// Call it
		LuaManager->OnTick();
		Commands->Start_Timer(pObj, this, LUAM_UPDATE_TICK_RATE, 1);
	}
}

////////////////////////////////////////////////////
ScriptRegistrant<LUA_INTERNAL_UPDATE> LUA_INTERNAL_UPDATE_Registrant("LUA_INTERNAL_UPDATE", "");