////////////////////////////////////////////////////
// CLuaEngineManager.h
//
// Purpose: Manages the Lua engine
//
// File History:
//	- 8/13/07 : File created - Kevin Kirst
////////////////////////////////////////////////////

#ifndef _CLUAENGINEMANAGER_H_
#define _CLUAENGINEMANAGER_H_

#include <string>
#include "scripts.h"

#ifndef GameObject
#define GameObject unsigned char
#endif //GameObject

#define LUAM_UPDATE_TICK_RATE	(0.01f)	// 100 ms

// Common script paths
#define SCRIPT_PATH_BASE	("Data\\Scripts\\")
#define SCRIPT_PATH_COMMON	("Data\\Scripts\\Common\\")
#define SCRIPT_PATH_MAPS	("Data\\Scripts\\Maps\\")
#define SCRIPT_PATH_ENTITIES	("Data\\Scripts\\Entities\\")

// Bring in Lua engine
#include "LuaEngine\\include\\LuaEngine.h"
#pragma comment (lib, "LuaEngine\\lib\\LuaEngine.lib")
#pragma comment (lib, "LuaEngine\\lib\\lua5.1.lib")

class CScriptObjectCommands;
class CScriptObjectLevel;
class CScriptObjectSystem;
class CScriptObjectPlayer;

////////////////////////////////////////////////////
class CLuaEngineManager
{
	IScriptSystem *m_pSS;

	// Update object (used to update tick)
	GameObject *m_pUpdateObject;

	// Last update tick
	unsigned long m_dwLastTick;

	// Current map dir (for script loading)
	std::string m_szCurrLevel;
	CScriptObjectCommands *m_pCommandsScriptObject;
	CScriptObjectLevel *m_pLevelScriptObject;
	CScriptObjectSystem *m_pSystemScriptObject;

	// Player map
	typedef std::map<int, CScriptObjectPlayer*> PlayerMap;
	PlayerMap m_PlayerSOMap;

	////////////////////////////////////////////////////
	// Constructor
	////////////////////////////////////////////////////
	CLuaEngineManager(void);
	CLuaEngineManager(CLuaEngineManager const&) {}
	CLuaEngineManager& operator =(CLuaEngineManager const&) {return *this;}

	// Single instance
	static CLuaEngineManager m_Instance;

	////////////////////////////////////////////////////
	// OnLuaError
	//
	// Purpose: Called when a script error occurs
	//
	// In:	szError - Error message
	////////////////////////////////////////////////////
	static void OnLuaError(char const* szError);

public:
	////////////////////////////////////////////////////
	// Destructor
	////////////////////////////////////////////////////
	virtual ~CLuaEngineManager(void);

	////////////////////////////////////////////////////
	// GetInstance
	//
	// Purpose: Return single instance of class
	////////////////////////////////////////////////////
	static CLuaEngineManager& GetInstance(void);

	////////////////////////////////////////////////////
	// Type-cast to Script System (for fast usage)
	////////////////////////////////////////////////////
	operator IScriptSystem*(void);

public:
	////////////////////////////////////////////////////
	// Initialize
	//
	// Purpose: Perform one-time initialization
	////////////////////////////////////////////////////
	void Initialize(void);

	////////////////////////////////////////////////////
	// GetScriptPath
	//
	// Purpose: Get the path + script file name
	//
	// In:	szScript - Script file name ("Script.lua")
	//
	// Out: szPath - Script path + file name ("Scripts\Script.lua")
	////////////////////////////////////////////////////
	void GetScriptPath(char const* szScript, char *szPath);
	std::string GetScriptPath(char const* szScript);

	////////////////////////////////////////////////////
	// OnLevelLoad
	//
	// Purpose: Call when a level has been loaded
	//
	// In:	szLevel - Name of the level that loaded
	////////////////////////////////////////////////////
	void OnLevelLoaded(char const* szLevel);

	////////////////////////////////////////////////////
	// OnObjectCreated
	//
	// Purpose: Call when an object is created, to load
	//	its script reference
	//
	// In:	pObj - Object that was created
	////////////////////////////////////////////////////
	void OnObjectCreated(GameObject *pObj);

	////////////////////////////////////////////////////
	// OnPurchase
	//
	// Purpose: Call when a purchase is made
	//
	// In:	pPurchaser - Who made the purchase
	//		nCost - How much it cost
	//		nPresetID - What was bought
	////////////////////////////////////////////////////
	void OnPurchase(GameObject *pPurchaser, unsigned int nCost, unsigned int nPresetID);

	////////////////////////////////////////////////////
	// OnTick
	//
	// Purpose: Called on update tick
	////////////////////////////////////////////////////
	void OnTick(void);

	////////////////////////////////////////////////////
	// OnChatMessage
	//
	// Purpose: Call when a chat message occurs
	//
	// In:	nID - Chat ID
	//		nType - Chat type
	//		szMsg - Chat message
	////////////////////////////////////////////////////
	void OnChatMessage(int nID, int nType, char const* szMsg);

	////////////////////////////////////////////////////
	// OnHostMessage
	//
	// Purpose: Call when a host message occurs
	//
	// In:	nType - Host message type (channel)
	//		szMsg - Host message
	////////////////////////////////////////////////////
	void OnHostMessage(int nType, char const* szMsg);

	////////////////////////////////////////////////////
	// OnPlayerJoined
	//
	// Purpose: Call when a player joins the server
	//
	// In:	nID - Player ID
	//
	// Note: Creates an entry in player map for player
	////////////////////////////////////////////////////
	void OnPlayerJoined(int nPlayerID);

	////////////////////////////////////////////////////
	// OnPlayerLeft
	//
	// Purpose: Call when a player leaves the server
	//
	// In:	nID - Player ID
	//
	// Note: Removes entry in player map for player
	////////////////////////////////////////////////////
	void OnPlayerLeft(int nPlayerID);
};

////////////////////////////////////////////////////
// Update Script
class LUA_INTERNAL_UPDATE : public ScriptImpClass
{
	void Created(GameObject *pObj);
	void Timer_Expired(GameObject *pObj, int nTimerID);
};

extern CLuaEngineManager *LuaManager;
extern IScriptSystem *Lua;

#endif //_CLUAENGINEMANAGER_H_