//////////////////////////////////////////////////////
// IScriptObject.h
//
// Purpose: Interface object
//	Describes a handler for interfacing to lua table
//	objects
//
// History:
//	- 5/17/07 : File created - Kevin Kirst
//////////////////////////////////////////////////////

#ifndef _ISCRIPTOBJECT_H_
#define _ISCRIPTOBJECT_H_

#include "IScriptSystem.h"

//////////////////////////////////////////////////////
struct lua_State;
struct IScriptObject
{
	friend struct IScriptSystem;
	friend struct IFunctionHandler;

	int _nRefCount;

public:
	//////////////////////////////////////////////////////
	// Constructor
	//////////////////////////////////////////////////////
	IScriptObject(void) : _nRefCount(1) {}

public:
	//////////////////////////////////////////////////////
	// Destructor
	//////////////////////////////////////////////////////
	virtual ~IScriptObject(void) {}

	//////////////////////////////////////////////////////
	// OpenTable
	//
	// Purpose: Used to open table internally
	//////////////////////////////////////////////////////
	virtual bool OpenTable(void) = 0;

	//////////////////////////////////////////////////////
	// CloseTable
	//
	// Purpose: Used to close table internally
	//////////////////////////////////////////////////////
	virtual void CloseTable(void) = 0;

	//////////////////////////////////////////////////////
	// Attach
	//
	// Purpose: Attaches the object to a lua state
	//
	// In:	pState - Lua state
	//		szName - Name of the table
	//////////////////////////////////////////////////////
	virtual void Attach(lua_State *pState, char const* szName) = 0;
	virtual void Attach(lua_State *pState, int nIndex) = 0;

	//////////////////////////////////////////////////////
	// AttachParent
	//
	// Purpose: Attaches the object to a parent object
	//
	// In:	pParent - Parent object
	//		szName - Name of the table
	//		nIndex - index of the table
	//////////////////////////////////////////////////////
	virtual void AttachParent(IScriptObject *pParent, char const* szName) = 0;
	virtual void AttachParent(IScriptObject *pParent, int nIndex) = 0;

	//////////////////////////////////////////////////////
	// Assign
	//
	// Purpose: Creates a table in the stack, assigns it
	//	to the name, and attatched it to this object
	//
	// In:	szName - Name of the table in Lua
	//////////////////////////////////////////////////////
	virtual void Assign(IScriptSystem *pState, char const* szName) = 0;

	//////////////////////////////////////////////////////
	// Reset
	//
	// Purpose: Resets script object
	//////////////////////////////////////////////////////
	virtual void Reset(void) = 0;

	//////////////////////////////////////////////////////
	// GetRef
	//
	// Purpose: Returns reference to script object
	//////////////////////////////////////////////////////
	IScriptObject *GetRef(void) { _nRefCount++; return this; };

	//////////////////////////////////////////////////////
	// Release
	//
	// Purpose: Release instance
	//////////////////////////////////////////////////////
	void Release(void)
	{
		if (_nRefCount > 0 && --_nRefCount == 0)
		{
			OnRelease();
			delete this;
		}
	}

	//////////////////////////////////////////////////////
	// OnRelease
	//
	// Purpose: Called when object is deleted
	//////////////////////////////////////////////////////
	virtual void OnRelease(void) = 0;

	//////////////////////////////////////////////////////
	// Write
	//
	// Purpose: Force writing of table
	//////////////////////////////////////////////////////
	virtual void Write(void) = 0;

	//////////////////////////////////////////////////////
	// Clone
	//
	// Purpose: Assimilates a table
	//
	// In:	pTbl - Table to clone
	//////////////////////////////////////////////////////
	virtual void Clone(IScriptObject *pTbl) = 0;

	//////////////////////////////////////////////////////
	// Clear
	//
	// Purpose: Clear the object
	//////////////////////////////////////////////////////
	virtual void Clear(void) = 0;

	//////////////////////////////////////////////////////
	// Erase
	//
	// Purpose: Erases the object (by setting it to NIL)
	//////////////////////////////////////////////////////
	virtual void Erase(void) = 0;

	//////////////////////////////////////////////////////
	// Size
	//
	// Purpose: Return number of elements in object array
	//////////////////////////////////////////////////////
	virtual int Size(void) = 0;

	//////////////////////////////////////////////////////
	// SetValue
	//
	// Purpose: Set a value in the object
	//
	// In:	szName - Name of value
	//		val - Value to set
	//////////////////////////////////////////////////////
	virtual void SetValue(char const* szName, int val) = 0;
	virtual void SetValue(char const* szName, size_t val) = 0;
	virtual void SetValue(char const* szName, float val) = 0;
	virtual void SetValue(char const* szName, bool val) = 0;
	virtual void SetValue(char const* szName, char const* val) = 0;
	virtual void SetValue(char const* szName, IScriptObject* val) = 0;
	virtual void SetValueToNull(char const* szName) = 0;

	//////////////////////////////////////////////////////
	// SetValueAt
	//
	// Purpose: Set value at index
	//
	// In:	index - Location of value
	//		val - Value to set
	//////////////////////////////////////////////////////
	virtual void SetValueAt(int index, int val) = 0;
	virtual void SetValueAt(int index, size_t val) = 0;
	virtual void SetValueAt(int index, float val) = 0;
	virtual void SetValueAt(int index, bool val) = 0;
	virtual void SetValueAt(int index, char const* val) = 0;
	virtual void SetValueAt(int index, IScriptObject* val) = 0;
	virtual void SetValueAtToNull(int index) = 0;

	//////////////////////////////////////////////////////
	// PushValue
	//
	// Purpose: Pushes a value to the end of the array
	//
	// In:	val - Value to push
	//////////////////////////////////////////////////////
	virtual void PushValue(int val) = 0;
	virtual void PushValue(size_t val) = 0;
	virtual void PushValue(float val) = 0;
	virtual void PushValue(bool val) = 0;
	virtual void PushValue(char const* val) = 0;
	virtual void PushValue(IScriptObject* val) = 0;
	virtual void PushNil(void) = 0;

	//////////////////////////////////////////////////////
	// GetValue
	//
	// Purpose: Return named value stored in object
	//
	// In:	szName - Name of value
	//
	// Out:	out - Value stored
	//
	// Returns: TRUE if value was retrieved
	//////////////////////////////////////////////////////
	virtual bool GetValue(char const* szName, int &out) = 0;
	virtual bool GetValue(char const* szName, size_t &out) = 0;
	virtual bool GetValue(char const* szName, float &out) = 0;
	virtual bool GetValue(char const* szName, bool &out) = 0;
	virtual bool GetValue(char const* szName, char * out, int len) = 0;
	virtual bool GetValue(char const* szName, std::string &out) = 0;
	virtual bool GetValue(char const* szName, IScriptObject* out) = 0;

	//////////////////////////////////////////////////////
	// GetValueAt
	//
	// Purpose: Returns value at index
	//
	// In:	index - Location of value
	//
	// Out:	out - Value stored
	//
	// Returns: TRUE if value was retrieved
	//////////////////////////////////////////////////////
	virtual bool GetValueAt(int index, int &out) = 0;
	virtual bool GetValueAt(int index, size_t &out) = 0;
	virtual bool GetValueAt(int index, float &out) = 0;
	virtual bool GetValueAt(int index, bool &out) = 0;
	virtual bool GetValueAt(int index, char*out, int len) = 0;
	virtual bool GetValueAt(int index, std::string &out) = 0;
	virtual bool GetValueAt(int index, IScriptObject* out) = 0;

	//////////////////////////////////////////////////////
	// HasValue
	//
	// Purpose: Returns TRUE if value is in table
	//
	// In:	szName - Name of value
	//////////////////////////////////////////////////////
	virtual bool HasValue(char const* szName) = 0;

	//////////////////////////////////////////////////////
	// GetValueType
	//
	// Purpose: Returns variable type
	//
	// In:	szName - Name of value
	//////////////////////////////////////////////////////
	virtual VarType GetValueType(char const* szName) = 0;
	virtual VarType GetAtType(int index) = 0;

	//////////////////////////////////////////////////////
	// RegisterFunction
	//
	// Purpose: Registers a C function to a Lua function
	//	contained in the table
	//
	// In:	szName - Name of the Lua function
	//		func - C function
	//////////////////////////////////////////////////////
	virtual void RegisterFunction(char const* szName, lua_CFunction func) = 0;
};


#endif //_ISCRIPTOBJECT_H_