//////////////////////////////////////////////////////
// IScriptSystem.h
//
// Purpose: Interface object
//	Describes a Lua Script System that handles a lua
//	object and script loading
//
// History:
//	- 5/17/07 : File created - Kevin Kirst
//////////////////////////////////////////////////////

#ifndef _ISCRIPTSYSTEM_H_
#define _ISCRIPTSYSTEM_H_

//////////////////////////////////////////////////////
enum VarType
{
	luaNull = 0,	// Null or Nil
	luaNumber,		// Interger or floating-point number
	luaString,		// String
	luaObject,		// Table
	luaFunction,	// Function
};

inline char const* GetVarTypeString(VarType v)
{
	switch (v)
	{
		case luaFunction: return "Function";
		case luaObject: return "Object";
		case luaString: return "String";
		case luaNumber: return "Number";
		case luaNull: return "Null";
		default: return "UNKNOWN";
	};
	return "ERROR";
};

typedef unsigned long USER_DATA;
typedef unsigned int UINT_PTR;
typedef int INT_PTR;

// Script error callback function signature
typedef void(*LUAENGINE_ERRORCALLBACK)(char const*);

//////////////////////////////////////////////////////
//////////////////////////////////////////////////////
struct IScriptObject;
struct IFunctionHandler;
struct IScriptSystem
{
	int _nRefCount;

protected:
	//////////////////////////////////////////////////////
	// Constructor
	//////////////////////////////////////////////////////
	IScriptSystem(void) : _nRefCount(1) {}

public:
	//////////////////////////////////////////////////////
	// Destructor
	//////////////////////////////////////////////////////
	virtual ~IScriptSystem(void) {}

	//////////////////////////////////////////////////////
	// GetRef
	//
	// Purpose: Returns reference to script system
	//////////////////////////////////////////////////////
	IScriptSystem *GetRef(void) { _nRefCount++; return this; };

	//////////////////////////////////////////////////////
	// Release
	//
	// Purpose: Release instance
	//////////////////////////////////////////////////////
	void Release(void)
	{
		if (_nRefCount > 0 && --_nRefCount == 0)
		{
			OnRelease();
			delete this;
		}
	}

	//////////////////////////////////////////////////////
	// OnRelease
	//
	// Purpose: Called when system is deleted
	//////////////////////////////////////////////////////
	virtual void OnRelease(void) = 0;

	//////////////////////////////////////////////////////
	// Initialize
	//
	// Purpose: Initialize the lua object
	//
	// In:	bSilent - TRUE to run in silent (no logging)
	//			mode
	//////////////////////////////////////////////////////
	virtual bool Initialize(bool bSilent = false) = 0;

	//////////////////////////////////////////////////////
	// RaiseErorr
	//
	// Purpose: Manually raise an error in the system
	//
	// In:	szError - Error message to print
	//		... - printf-like parameters
	//
	// Note: Pass 'NULL' for szError if message is on
	//	the top of the stack!
	//////////////////////////////////////////////////////
	virtual void RaiseError(char const* szError = NULL, ...) = 0;

	//////////////////////////////////////////////////////
	// GetTop
	//
	// Purpose: Returns top stack index
	//////////////////////////////////////////////////////
	virtual int GetTop(void) = 0;

	//////////////////////////////////////////////////////
	// GetState
	//
	// Purpose: Returns Lua state
	//////////////////////////////////////////////////////
	virtual lua_State *GetState(void) const = 0;

	//////////////////////////////////////////////////////
	// IsAlive
	//
	// Purpose: Returns true if Lua state is valid
	//////////////////////////////////////////////////////
	virtual bool IsAlive(void) const = 0;

	//////////////////////////////////////////////////////
	// ForceGarbageCollect
	//
	// Purpose: Force the stack to flush and remove any
	//	unused table objects
	//////////////////////////////////////////////////////
	virtual void ForceGarbageCollect(void) = 0;

	//////////////////////////////////////////////////////
	// GetMemoryUsage
	//
	// Purpose: Returns how much memory is in use by Lua
	//	right now.
	//
	// Note: In debug mode, also stores if max
	//////////////////////////////////////////////////////
	virtual int GetMemoryUsage(void) = 0;

	//////////////////////////////////////////////////////
	// ExecuteCommand
	//
	// Purpose: Executes a raw chunk
	//
	// In:	szChunk - Chunk to execute
	//////////////////////////////////////////////////////
	virtual bool ExecuteCommand(char const* szChunk, char const* szChunkName = "chunk") = 0;

	//////////////////////////////////////////////////////
	// LoadScript
	//
	// Purpose: Loads a script if it hasn't already been
	//	loaded.  Otherwise, ignores
	//
	// In:	szFile - Script to load
	//
	// Returns: TRUE if script was loaded without error
	//////////////////////////////////////////////////////
	virtual bool LoadScript(char const* szFile) = 0;

	//////////////////////////////////////////////////////
	// UnloadScript
	//
	// Purpose: Unloads a script file
	//
	// In:	szFile - Script to load
	//////////////////////////////////////////////////////
	virtual void UnloadScript(char const* szFile) = 0;

	//////////////////////////////////////////////////////
	// UnloadScripts
	//
	// Purpose: Unloads all script files
	//////////////////////////////////////////////////////
	virtual void UnloadScripts(void) = 0;

	//////////////////////////////////////////////////////
	// ReloadScript
	//
	// Purpose: Forces a reload of an already loaded
	//	script.  If not already loaded, is loaded.
	//
	// In:	szFile - Script to (re)load
	//////////////////////////////////////////////////////
	virtual bool ReloadScript(char const* szFile) = 0;

	//////////////////////////////////////////////////////
	// ReloadScripts
	//
	// Purpose: Forces a reload on all loaded scripts
	//////////////////////////////////////////////////////
	virtual bool ReloadScripts(void) = 0;

	//////////////////////////////////////////////////////
	// CreateEmptyObject
	//
	// Purpose: Returns an empty object
	//////////////////////////////////////////////////////
	virtual IScriptObject *CreateEmptyObject(void) = 0;

	//////////////////////////////////////////////////////
	// CreateObject
	//
	// Purpose: Returns a nameless object
	//////////////////////////////////////////////////////
	virtual IScriptObject *CreateObject(void) = 0;

	//////////////////////////////////////////////////////
	// CreateGlobalObject
	//
	// Purpose: Creates an empty global object
	//
	// In:	szName - Name of the global
	//////////////////////////////////////////////////////
	virtual IScriptObject *CreateGlobalObject(char const* szName) = 0;

	//////////////////////////////////////////////////////
	// CreateRawObject
	//
	// Purpose: Just return a script object off the heap
	//////////////////////////////////////////////////////
	virtual IScriptObject *CreateRawObject(void) const = 0;

	//////////////////////////////////////////////////////
	// GetFunctionHelper
	//
	// Purpose: Returns a function helper
	//////////////////////////////////////////////////////
	virtual IFunctionHandler *GetFunctionHelper(void) = 0;

	//////////////////////////////////////////////////////
	// SetGlobal
	//
	// Purpose: Sets a global value
	//
	// In:	szName - Name of the global
	//		val - Value of the global
	//////////////////////////////////////////////////////
	virtual void SetGlobal(char const* szName, int val) = 0;
	virtual void SetGlobal(char const* szName, size_t val) = 0;
	virtual void SetGlobal(char const* szName, float val) = 0;
	virtual void SetGlobal(char const* szName, bool val) = 0;
	virtual void SetGlobal(char const* szName, char const* val) = 0;
	virtual void SetGlobalToNull(char const* szName) = 0;

	//////////////////////////////////////////////////////
	// IsGlobalDefined
	//
	// Purpose: Returns TRUE if global value is defined
	//
	// In:	szName - Name of the global
	//////////////////////////////////////////////////////
	virtual bool IsGlobalDefined(char const* szName) = 0;

	//////////////////////////////////////////////////////
	// GetGlobal
	//
	// Purpose: Returns the value of a global
	//
	// In:	szName - Name of the global
	//
	// Out: out - Stored value
	//
	// Returns: TRUE on success
	//////////////////////////////////////////////////////
	virtual bool GetGlobal(char const* szName, int &out) = 0;
	virtual bool GetGlobal(char const* szName, size_t &out) = 0;
	virtual bool GetGlobal(char const* szName, float &out) = 0;
	virtual bool GetGlobal(char const* szName, bool &out) = 0;
	virtual bool GetGlobal(char const* szName, char*out, int len) = 0; // Last character set to NULL terminator
	virtual bool GetGlobal(char const* szName, std::string &out) = 0;
	virtual bool GetGlobal(char const* szName, IScriptObject *out) = 0;

	//////////////////////////////////////////////////////
	// GetGlobalType
	//
	// Purpose: Returns global variable type
	//
	// In:	szName - Name of global
	//////////////////////////////////////////////////////
	virtual VarType GetGlobalType(char const* szName) = 0;

	//////////////////////////////////////////////////////
	// BeginCall
	//
	// Purpose: Begin a function call
	//
	// In:	szTable - Table where function is contained
	//		szName - Name of the function
	//
	// Returns: TRUE if function was found
	//////////////////////////////////////////////////////
	virtual bool BeginCall(char const* szName) = 0;
	virtual bool BeginCall(char const* szTable, char const* szName) = 0;

	//////////////////////////////////////////////////////
	// PushArgument
	//
	// Purpose: Push an argument into a function call
	//
	// In:	arg - Argument to be pushed in
	//////////////////////////////////////////////////////
	virtual void PushArgument(int arg) = 0;
	virtual void PushArgument(size_t arg) = 0;
	virtual void PushArgument(float arg) = 0;
	virtual void PushArgument(bool arg) = 0;
	virtual void PushArgument(char const* arg) = 0;
	virtual void PushArgument(std::string const& arg) = 0;
	virtual void PushArgument(IScriptObject* arg) = 0;
	virtual void PushNullArgument(void) = 0;

	//////////////////////////////////////////////////////
	// EndCall
	//
	// Purpose: Executes the function call with pushed
	//	arguments and gets returned values
	//
	// Out: ret - Returned value
	//////////////////////////////////////////////////////
	virtual void EndCall(void) = 0;
	virtual bool EndCall(int &ret) = 0;
	virtual bool EndCall(size_t &ret) = 0;
	virtual bool EndCall(float &ret) = 0;
	virtual bool EndCall(bool &ret) = 0;
	virtual bool EndCall(char*ret, int len) = 0;
	virtual bool EndCall(std::string &ret) = 0;
	virtual bool EndCall(IScriptObject *ret) = 0;

	//////////////////////////////////////////////////////
	// RegisterGlobalFunction
	//
	// Purpose: Registers a global C function in the script
	//
	// In:	szName - Name of the function (in Lua)
	//		func - Lua function
	//////////////////////////////////////////////////////
	virtual void RegisterGlobalFunction(char const* szName, lua_CFunction func) = 0;
};


#endif //_ISCRIPTSYSTEM_H_