//////////////////////////////////////////////////////
// IScriptVar.h
//
// Purpose: Interface objects
//	Describes a Lua global variable wrapper
//
// History:
//	- 5/17/07 : File created - Kevin Kirst
//////////////////////////////////////////////////////

#ifndef _ISCRIPTVAR_H_
#define _ISCRIPTVAR_H_

#include "IScriptSystem.h"

struct IScriptVar
{
	int _nRefCount;

protected:
	//////////////////////////////////////////////////////
	// Constructor
	//////////////////////////////////////////////////////
	IScriptVar(void) : _nRefCount(1) {}

public:
	//////////////////////////////////////////////////////
	// Destructor
	//////////////////////////////////////////////////////
	virtual ~IScriptVar(void) {}

	//////////////////////////////////////////////////////
	// GetRef
	//
	// Purpose: Returns reference to script system
	//////////////////////////////////////////////////////
	IScriptVar *GetRef(void) { _nRefCount++; return this; };

	//////////////////////////////////////////////////////
	// Release
	//
	// Purpose: Release instance
	//////////////////////////////////////////////////////
	void Release(void)
	{
		if (_nRefCount > 0 && --_nRefCount == 0)
		{
			OnRelease();
			delete this;
		}
	}

	//////////////////////////////////////////////////////
	// OnRelease
	//
	// Purpose: Called when system is deleted
	//////////////////////////////////////////////////////
	virtual void OnRelease(void) = 0;

	//////////////////////////////////////////////////////
	// GetScriptSystem
	//
	// Purpose: Returns script system where variable is in
	//////////////////////////////////////////////////////
	virtual struct IScriptSystem* GetScriptSystem(void) const = 0;

	//////////////////////////////////////////////////////
	// GetName
	//
	// Purpose: Returns name of variable
	//////////////////////////////////////////////////////
	virtual char const* GetName(void) const = 0;

	//////////////////////////////////////////////////////
	// Initialize
	//
	// Purpose: Direct the object to the Lua variable
	//
	// In:	pScriptSystem - Lua script system
	//		szName - Global variable name
	//		defValue - Default value (if global is not defined)
	//////////////////////////////////////////////////////
	virtual void Initialize(IScriptSystem *pScriptSystem, char const* szName, int defValue) = 0;
	virtual void Initialize(IScriptSystem *pScriptSystem, char const* szName, float defValue) = 0;
	virtual void Initialize(IScriptSystem *pScriptSystem, char const* szName, bool defValue) = 0;
	virtual void Initialize(IScriptSystem *pScriptSystem, char const* szName, char const* defValue) = 0;
	virtual void InitializeToNull(IScriptSystem *pScriptSystem, char const* szName) = 0;

	//////////////////////////////////////////////////////
	// GetValue<type>
	//
	// Purpose: Gets the value of the variable
	//////////////////////////////////////////////////////
	virtual int GetValueInt(void) = 0;
	virtual float GetValueFloat(void) = 0;
	virtual bool GetValueBool(void) = 0;
	virtual void GetValueString(char *str, int nLen) = 0;
	virtual void GetValueString(std::string &str) = 0;

	//////////////////////////////////////////////////////
	// SetValue
	//
	// Purpose: Sets the value of the variable
	//////////////////////////////////////////////////////
	virtual void SetValue(int val) = 0;
	virtual void SetValue(float val) = 0;
	virtual void SetValue(bool val) = 0;
	virtual void SetValue(char const* val) = 0;
};

#endif //_ISCRIPTVAR_H_