//////////////////////////////////////////////////////
// IScriptableClass.h
//
// Purpose: Inherit interface to make a class scriptable
//	so that it can expose functions and member variables
//	to a script object
//
// History:
//	- 5/17/07 : File created - Kevin Kirst
//////////////////////////////////////////////////////

#ifndef _ISCRIPTABLECLASS_H_
#define _ISCRIPTABLECLASS_H_

#include "IScriptSystem.h"

#pragma warning (disable : 4311) // Shown for m_nPropertyBase assignment truncation
#pragma warning (disable : 4312) // Shown for m_nPropertyBase base+offset location access

// Property types
enum PROPERTY_TYPE
{
	PROPERTY_TYPE_INT = 1,
	PROPERTY_TYPE_UINT,
	PROPERTY_TYPE_SHORT,
	PROPERTY_TYPE_USHORT,
	PROPERTY_TYPE_CHAR,
	PROPERTY_TYPE_UCHAR,
	PROPERTY_TYPE_BOOL,
	PROPERTY_TYPE_FLOAT,
	PROPERTY_TYPE_DOUBLE,
	PROPERTY_TYPE_STRING,
	PROPERTY_TYPE_OBJECT,
};

#define USE_INTERNAL_UID (0xFFFFFFFFFFFFFFFF)

template <class T>
struct IScriptableClass
{
	typedef int(T::*MemberFunc)(IFunctionHandler*);
	IScriptSystem *m_pSS;
	IScriptObject *m_pSO;
	UINT_PTR m_nPropertyBase;

	struct Property
	{
		Property(void)
		{
			offsetloc = 0;
		}
		Property(Property const& obj)
		{
			type = obj.type;
			offsetloc = obj.offsetloc;
		}
		PROPERTY_TYPE type;
		unsigned int offsetloc;
	};

	typedef std::map<std::string, Property> PropertyMap;
	typedef std::map<std::string, MemberFunc> FuncMap;
	typedef std::map<void*, T*> ClassMap;
	static FuncMap m_FuncMap;
	static PropertyMap m_PropMap;
	static ClassMap m_ClassMap;
	static IScriptObject *m_pTemplate;
	static IScriptSystem *m_pTemplateSystem;

public:
	//////////////////////////////////////////////////////
	// Constructor
	//////////////////////////////////////////////////////
	IScriptableClass(void)
	{
		m_pSS = NULL;
		m_pSO = NULL;
		m_nPropertyBase = 0;
	}

	//////////////////////////////////////////////////////
	// Destructor
	//////////////////////////////////////////////////////
	virtual ~IScriptableClass(void)
	{
		Shutdown();
	}

	//////////////////////////////////////////////////////
	// GetScriptSystem
	//
	// Purpose: Return internal script system
	//////////////////////////////////////////////////////
	inline IScriptSystem *GetScriptSystem(void) const { return m_pSS; }

	//////////////////////////////////////////////////////
	// GetScriptObject
	//
	// Purpose: Return internal script object
	//////////////////////////////////////////////////////
	inline IScriptObject *GetScriptObject(void) const { return m_pSO; }

public:
	//////////////////////////////////////////////////////
	// Init
	//
	// Purpose: Initialize the class to a script system
	//
	// In:	pScriptSystem - Script system
	//		szName - Name of script object to create
	//		pParent - Parent class
	//		bUnique - TRUE to make unique table
	//////////////////////////////////////////////////////
	void Init(IScriptSystem *pScriptSystem, char const* szName, T* pParent, bool bUnique = true)
	{
		if (NULL == pScriptSystem) return;
		if (NULL != m_pSO)
		{
			m_pSO->Erase();
			m_pSO->Release();
		}
		if (NULL != m_pSS) m_pSS->Release();

		m_pSS = pScriptSystem->GetRef();
		static unsigned long long _InitID = 0;
		char szTableName[64];
		if (bUnique)
			_snprintf(szTableName, 64, "%s%lu", szName, _InitID++);
		else
			_snprintf(szTableName, 64, "%s", szName);
		m_pSS->SetGlobalToNull(szTableName);
		char szPropertyMetaTableName[64];
		sprintf(szPropertyMetaTableName, "%s.CNT", szTableName);

		// Create metatable for entity property "table"
		lua_State *pState = m_pSS->GetState();
		luaL_newmetatable(pState, szPropertyMetaTableName);
		static const struct luaL_Reg _entcntmeta[] =
		{
			{"__index", T::GetPropCallback},
			{"__newindex", T::SetPropCallback},
			{NULL, NULL},
		};
		luaL_register(pState, NULL, _entcntmeta);
		lua_pop(pState, 1);

		// Create the script object
		m_pSO = m_pSS->CreateRawObject();
		m_pSO->Assign(pScriptSystem, szTableName);
		
		// Clone global table and template table
		IScriptObject *pClass = pScriptSystem->CreateEmptyObject();
		pScriptSystem->GetGlobal(szName, pClass);
		m_pSO->Clone(pClass);
		m_pSO->Clone(m_pTemplate);

		// Create the properties userdata table
		if (m_pSO->OpenTable())
		{
			lua_pushstring(pState, "cnt");
			void* ud = lua_newuserdata(pState, 8);
			m_ClassMap[ud] = pParent;
			luaL_getmetatable(pState, szPropertyMetaTableName);
			lua_setmetatable(pState, -2);
			lua_settable(pState, -3);
			m_pSO->CloseTable();
		}

		// Register it
		m_pSO->Write();

		// Clean up
		pClass->Release();
	};

	//////////////////////////////////////////////////////
	// Shutdown
	//
	// Purpose: Clean up
	//////////////////////////////////////////////////////
	void Shutdown(void)
	{
		if (NULL != m_pSO)
		{
			m_pSO->Erase();
			m_pSO->Release();
			m_pSO = NULL;
		}
		if (NULL != m_pSS)
		{
			m_pSS->Release();
			m_pSS = NULL;
		}
	}

	//////////////////////////////////////////////////////
	// InitializeTemplate
	//
	// Purpose: Set up template table for derived classes
	//
	// In:	pSS - Script system to use
	//	szTemplateName - Unique name for template table
	//////////////////////////////////////////////////////
	static void InitializeTemplate(IScriptSystem *pSS, char const* szTemplateName)
	{
		if (NULL == m_pTemplateSystem)
		{
			m_pTemplateSystem = pSS->GetRef();
		}
		if (NULL != m_pTemplateSystem && NULL == m_pTemplate)
		{
			m_pTemplate = m_pTemplateSystem->CreateRawObject();
			char szTableName[64];
			_snprintf(szTableName, 64, "__TMLTBL_%s__", szTemplateName);
			m_pTemplate->Assign(m_pTemplateSystem, szTableName);
		}
	}

	//////////////////////////////////////////////////////
	// ReleaseTemplate
	//
	// Purpose: Release template table and clean up
	//////////////////////////////////////////////////////
	static void ReleaseTemplate(void)
	{
		if (NULL != m_pTemplate)
		{
			m_pTemplate->Release();
			m_pTemplate = NULL;
		}
		if (NULL != m_pTemplateSystem)
		{
			m_pTemplateSystem->Release();
			m_pTemplateSystem = NULL;
		}
	}

	//////////////////////////////////////////////////////
	// RegisterFunction
	//
	// Purpose: Registers a function to the script object
	//
	// In:	szName - Name of function in Lua
	//		func - Member function
	//////////////////////////////////////////////////////
	static void RegisterFunction(char const* szName, MemberFunc func)
	{
		if (NULL == m_pTemplateSystem || NULL == m_pTemplate) return;

		// Register it
		m_pTemplate->RegisterFunction(szName, T::FuncCallback);

		// Add it to function map
		m_FuncMap[szName] = func;
	};

	//////////////////////////////////////////////////////
	// EnablePropertyMapping
	//
	// Purpose: Sets up property mapping for instance
	//
	// In:	base - Base address of class
	//////////////////////////////////////////////////////
	bool EnablePropertyMapping(void* base)
	{
		m_nPropertyBase = (UINT_PTR)base;
		return true;
	}

	//////////////////////////////////////////////////////
	// RegisterProperty
	//
	// Purpose: Registers a member variable to the
	//	property table in the script object
	//
	// In:	szName - Name of the property in Lua
	//		type - Property type
	//		offset - Offset into class for member variable
	//////////////////////////////////////////////////////
	static void RegisterProperty(char const* szName, PROPERTY_TYPE type, int offset)
	{
		if (NULL == m_pTemplateSystem) return;

		// Create a new property entry for it
		Property prop;
		prop.type = type;
		prop.offsetloc = offset;
		m_PropMap[szName] = prop;
	};

	//////////////////////////////////////////////////////
	// GetPropCallback
	//
	// Purpose: Lua get property callback router
	//
	// In:	pState - Lua state object
	//////////////////////////////////////////////////////
	static int GetPropCallback(lua_State *pState)
	{
		// Get user data (cnt @ -2)
		void *ud = lua_touserdata(pState, -2);
		if (NULL == ud) return 0;

		// Get class by user data
		if (m_ClassMap.find(ud) == m_ClassMap.end()) return 0;
		T* pClass = m_ClassMap[ud];
		if (NULL == pClass) return 0;

		// Get key
		if (!lua_isstring(pState, -1)) return 0;
		char const* szKey = lua_tostring(pState, -1);

		// Get property entry
		if (m_PropMap.find(szKey) == m_PropMap.end()) return 0;
		Property &p = m_PropMap[szKey];

		// Based on type, push onto stack
		switch (p.type)
		{
			case PROPERTY_TYPE_INT:
			{
				int val = (int)*((int*)(pClass->m_nPropertyBase+p.offsetloc));
				lua_pushinteger(pState, val);
				return 1;
			};
			case PROPERTY_TYPE_UINT:
			{
				unsigned int val = (unsigned int)*((unsigned int*)(pClass->m_nPropertyBase+p.offsetloc));
				lua_pushinteger(pState, val);
				return 1;
			};
			case PROPERTY_TYPE_SHORT:
			{
				short val = (short)*((short*)(pClass->m_nPropertyBase+p.offsetloc));
				lua_pushinteger(pState, val);
				return 1;
			};
			case PROPERTY_TYPE_USHORT:
			{
				unsigned short val = (unsigned short)*((unsigned short*)(pClass->m_nPropertyBase+p.offsetloc));
				lua_pushinteger(pState, val);
				return 1;
			};
			case PROPERTY_TYPE_CHAR:
			{
				char val = (char)*((char*)(pClass->m_nPropertyBase+p.offsetloc));
				lua_pushinteger(pState, val);
				return 1;
			};
			case PROPERTY_TYPE_UCHAR:
			{
				unsigned char val = (unsigned char)*((unsigned char*)(pClass->m_nPropertyBase+p.offsetloc));
				lua_pushinteger(pState, val);
				return 1;
			};
			case PROPERTY_TYPE_BOOL:
			{
				bool val = (bool)*((bool*)(pClass->m_nPropertyBase+p.offsetloc));
				lua_pushinteger(pState, val==true?1:0);
				return 1;
			};
			case PROPERTY_TYPE_FLOAT:
			{
				float val = float(*((float*)(pClass->m_nPropertyBase+p.offsetloc)));
				lua_pushnumber(pState, val);
				return 1;
			};
			case PROPERTY_TYPE_DOUBLE:
			{
				double val = (double)*((double*)(pClass->m_nPropertyBase+p.offsetloc));
				lua_pushnumber(pState, val);
				return 1;
			};
			case PROPERTY_TYPE_STRING:
			{
				char* val = (char*)(pClass->m_nPropertyBase+p.offsetloc);
				lua_pushstring(pState, val);
				return 1;
			};
			case PROPERTY_TYPE_OBJECT:
			{
				return 0;
			};
		};

		return 0;
	};

	//////////////////////////////////////////////////////
	// SetPropCallback
	//
	// Purpose: Lua set property callback router
	//
	// In:	pState - Lua state object
	//////////////////////////////////////////////////////
	static int SetPropCallback(lua_State *pState)
	{
		// Get user data (cnt @ -3)
		void *ud = lua_touserdata(pState, -3);
		if (NULL == ud) return 0;

		// Get class by user data
		if (m_ClassMap.find(ud) == m_ClassMap.end()) return 0;
		T* pClass = m_ClassMap[ud];
		if (NULL == pClass) return 0;

		// Get key
		if (!lua_isstring(pState, -2)) return 0;
		char const* szKey = lua_tostring(pState, -2);

		// Get property entry
		if (m_PropMap.find(szKey) == m_PropMap.end()) return 0;
		Property &p = m_PropMap[szKey];

		// Based on type, push onto stack
		switch (p.type)
		{
			case PROPERTY_TYPE_INT:
			{
				if (!lua_isnumber(pState, -1)) return 0; // Invalid type
				int val = (int)lua_tonumber(pState, -1);
				*((int*)(pClass->m_nPropertyBase+p.offsetloc)) = val;
				return 0;
			};
			case PROPERTY_TYPE_UINT:
			{
				if (!lua_isnumber(pState, -1)) return 0; // Invalid type
				unsigned int val = (unsigned int)lua_tonumber(pState, -1);
				*((unsigned int*)(pClass->m_nPropertyBase+p.offsetloc)) = val;
				return 0;
			};
			case PROPERTY_TYPE_SHORT:
			{
				if (!lua_isnumber(pState, -1)) return 0; // Invalid type
				short val = (short)lua_tonumber(pState, -1);
				*((short*)(pClass->m_nPropertyBase+p.offsetloc)) = val;
				return 0;
			};
			case PROPERTY_TYPE_USHORT:
			{
				if (!lua_isnumber(pState, -1)) return 0; // Invalid type
				unsigned short val = (unsigned short)lua_tonumber(pState, -1);
				*((unsigned short*)(pClass->m_nPropertyBase+p.offsetloc)) = val;
				return 0;
			};
			case PROPERTY_TYPE_CHAR:
			{
				if (!lua_isnumber(pState, -1)) return 0; // Invalid type
				char val = (char)lua_tonumber(pState, -1);
				*((char*)(pClass->m_nPropertyBase+p.offsetloc)) = val;
				return 0;
			};
			case PROPERTY_TYPE_UCHAR:
			{
				if (!lua_isnumber(pState, -1)) return 0; // Invalid type
				unsigned char val = (unsigned char)lua_tonumber(pState, -1);
				*((unsigned char*)(pClass->m_nPropertyBase+p.offsetloc)) = val;
				return 0;
			};
			case PROPERTY_TYPE_BOOL:
			{
				if (!lua_isnumber(pState, -1)) return 0; // Invalid type
				bool val = (int)lua_tonumber(pState, -1)==0?false:true;
				*((bool*)(pClass->m_nPropertyBase+p.offsetloc)) = val;
				return 0;
			};
			case PROPERTY_TYPE_FLOAT:
			{
				if (!lua_isnumber(pState, -1)) return 0; // Invalid type
				float val = float(lua_tonumber(pState, -1));
				*((float*)(pClass->m_nPropertyBase+p.offsetloc)) = val;
				return 0;
			};
			case PROPERTY_TYPE_DOUBLE:
			{
				if (!lua_isnumber(pState, -1)) return 0; // Invalid type
				double val = (double)lua_tonumber(pState, -1);
				*((double*)(pClass->m_nPropertyBase+p.offsetloc)) = val;
				return 0;
			};
			case PROPERTY_TYPE_STRING:
			{
				if (!lua_isstring(pState, -1)) return 0; // Invalid type
				char const* val = lua_tostring(pState, -1);
				strcpy((char*)(pClass->m_nPropertyBase+p.offsetloc), val);
				return 0;
			};
			case PROPERTY_TYPE_OBJECT:
			{
				if (!lua_istable(pState, -1)) return 0; // Invalid type
				return 0;
			};
		};

		return 0;
	};

	//////////////////////////////////////////////////////
	// FuncCallback
	//
	// Purpose: Lua function callback router
	//
	// In:	pState - Lua state object
	//////////////////////////////////////////////////////
	static int FuncCallback(lua_State *pState)
	{
		// Get the name of the function that was called
		lua_Debug ar;
		if (!lua_getstack(pState, 0, &ar)) return 0;
		lua_getinfo(pState, "n", &ar);

		// Get cnt from self to get calling class
		T* pClass = NULL;
		if (lua_istable(pState, 1))
		{
			lua_pushvalue(pState, 1);
			lua_pushstring(pState, "cnt");
			lua_gettable(pState, -2);
			void *ud = lua_touserdata(pState, -1);
			if (NULL != ud)
			{
				// Get class by user data
				pClass = m_ClassMap[ud];
			}
			lua_pop(pState, 2); // user data and cloned self
		}
		if (NULL == pClass) return 0;
		
		// Find it in the map
		for (FuncMap::iterator itI = m_FuncMap.begin(); itI != m_FuncMap.end(); itI++)
		{
			if (itI->first == ar.name)
			{
				// Setup the function handler now
				IFunctionHandler *functionhandler = m_pTemplateSystem->GetFunctionHelper();
				functionhandler->_Attach(pState);

				// Call it
				return (pClass->*(itI->second))(functionhandler);
			}
		}

		// Could not find it in the map!
		return 0;
	};
};

//////////////////////////////////////////////////////
//////////////////////////////////////////////////////

// Call this macro on all classes you want to register!
#define MAKE_SCRIPTABLE(_class) \
	IScriptableClass<_class>::FuncMap IScriptableClass<_class>::m_FuncMap; \
	IScriptableClass<_class>::PropertyMap IScriptableClass<_class>::m_PropMap; \
	IScriptableClass<_class>::ClassMap IScriptableClass<_class>::m_ClassMap; \
	IScriptObject *IScriptableClass<_class>::m_pTemplate = NULL; \
	IScriptSystem *IScriptableClass<_class>::m_pTemplateSystem = NULL;


// Helper macro to check argument count
#define CHECK_ARGUMENTS(num) \
	if (pH->GetArgumentCount() < num) \
	{ \
		GetScriptSystem()->RaiseError("expected %d arguments, got %d", num, pH->GetArgumentCount()); \
	}

// Helper macro to check argument type
#define CHECK_ARGUMENT_TYPE(loc, type) \
	if (pH->GetArgumentType(loc) != type) \
	{ \
		GetScriptSystem()->RaiseError("invalid argument type %d (got %s, expected %s)", loc, GetVarTypeString(pH->GetArgumentType(loc)), GetVarTypeString(type)); \
	}



#endif //_ISCRIPTABLECLASS_H_